/*
        Thymio-II Bootloader

        Copyright (C) 2011 Philippe Retornaz <philippe dot retornaz at epfl dot ch>,
        Mobots group (http://mobots.epfl.ch), Robotics system laboratory (http://lsro.epfl.ch)
        EPFL Ecole polytechnique federale de Lausanne (http://www.epfl.ch)

        See authors.txt for more details about other contributors.

        This program is free software: you can redistribute it and/or modify
        it under the terms of the GNU Lesser General Public License as published
        by the Free Software Foundation, version 3 of the License.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU Lesser General Public License for more details.

        You should have received a copy of the GNU Lesser General Public License
        along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <p24fxxxx.h>

#include <usb/usb.h>
#include <usb/usb_device.h>
#include "usb_function_cdc.h"

_CONFIG1 (0x3F69)
_CONFIG2 (0x134D)
_CONFIG3 (0x6000)

#define BASE_TIMEOUT 40000
#define EXTENDED_TIMEOUT 9

#define CHARGE_500MA  _LATF0

static int flash_in_progress;

static int flash_ptr;
static int flash_page;
static unsigned long flash_buffer[512];
unsigned int usb_ptr;

static unsigned char __attribute((aligned(2))) usb_buffer[100]; // Should be long enough for the maxium size of an aseba packet + 64.

int do_flash_page(unsigned int page_hiptr, unsigned int page_lowptr, unsigned long * page_buf);
int read_page(unsigned int page_hiptr, unsigned int page_lowptr, unsigned long * page_buf);

void __attribute((noreturn)) goto_userspace(void);

static int got_packet;
#define STATUS_IDLE 0
#define STATUS_DATA 1
int status;
int read_status;

extern void _CODE_BASE(void); // Don't ask ...

void _reset(void);
void _USB1Interrupt(void);

//#define ENABLE_READ_CMD


#define ASEBA_CMD_RESET 0x8000 				// Exit bootloader
#define ASEBA_CMD_READ_PAGE 0x8001              // ask the bootloader to read a page and send it back
#define ASEBA_CMD_WRITE_PAGE 0x8002     // ask the bootloader to write a page
#define ASEBA_PUSH_DESCRIPTION 0x8004   // message containing bootloader description
#define ASEBA_PUSH_PAGE_DATA 0x8005             // message containing data to host
#define ASEBA_PUSH_ACK 0x8006                   // message containing return value
#define ASEBA_MESSAGE_DISCONNECTED 0x9004


#define ASEBA_ANSWER_OK 0x0
#define ASEBA_ANSWER_INVALID_SIZE 0x1
#define ASEBA_ANSWER_PROGRAM_FAILED 0x2
#define ASEBA_ANSWER_NOT_PROGRAMMING 0x3
#define ASEBA_ANSWER_INVALID_VALUE 0x4

static unsigned char __attribute__((aligned(2))) data_to_send[64];
static unsigned char len;
#define barrier() asm __volatile__ ("" : : : "memory")

unsigned char AsebaTxReady(unsigned char *data) {
	int ret = len;
	if(len) 
		memcpy(data,data_to_send,len);
	len = 0;
#ifdef ENABLE_READ_CMD
	if(read_status == STATUS_DATA) {
		// fill up next datas ... 
		if(2048 - flash_ptr * 4 > 64) {
			memcpy(data_to_send, &flash_buffer[flash_ptr], 64);
			flash_ptr += 16;
			len = 64;
		} else {
			len = 2048 - flash_ptr * 4;
			if(len)
				memcpy(data_to_send, &flash_buffer[flash_ptr], len);
			read_status = STATUS_IDLE; // Next packet is the last one ...
		}
	}
#endif
	return ret;
}

static void send_ack(int d) {
// TODO FIXME DEBUG HACK: We should check that no pending packets are in the buffer ?!
	int * p = (int *) data_to_send;
	*p++ = 2;
	*p++ = 1;
	*p++ = ASEBA_PUSH_ACK;
	*p++ = d;
	len = 8;
	USBCDCKickTx();
}

static void send_noprog(int d) {
// TODO FIXME DEBUG HACK: We should check that no pending packets are in the buffer ?!
	int * p = (int *) data_to_send;
	*p++ = 4;
	*p++ = 1;
	*p++ = ASEBA_PUSH_ACK;
	*p++ = ASEBA_ANSWER_PROGRAM_FAILED;
	*p++ = d;
	len = 10;
	USBCDCKickTx();
}


static int AsebaValidPacket(void) {
	// Sanity check ... emiting ID should be "Host" == 0
	if(*(((unsigned int *) usb_buffer) + 1) != 0) 
		return 1;
		
	switch(*(((unsigned int *) usb_buffer) + 2)) {
		case ASEBA_CMD_RESET:
			flash_in_progress = 0; // always obey the host ...
			
			// Flag to tell the main() loop "go to userspace"
			got_packet = -1;
			return 1;
			
		case ASEBA_CMD_WRITE_PAGE:
			flash_page = *(((unsigned int *) usb_buffer) + 4);
		
			// Everything else is data ... 
			unsigned int left =  usb_ptr - *((unsigned int *) usb_buffer) - 6;
			status = STATUS_DATA;
			flash_in_progress = 1;
			flash_ptr = 0;
			if(left) {
				memcpy(flash_buffer, &usb_buffer[usb_ptr - left], left);
				flash_ptr += left;
				usb_ptr = 0;
			}
			return 0;
#ifdef ENABLE_READ_CMD
		case ASEBA_CMD_READ_PAGE:
			flash_page = *(((unsigned int *) usb_buffer) + 4);
			flash_ptr = 64/4;
			read_status = STATUS_DATA;
			read_page(flash_page >> 6, flash_page  << 10,flash_buffer);
			memcpy(data_to_send, flash_buffer, 64);
			len = 64;
			USBCDCKickTx();
			return 1;
#endif
	}
	return 1;
}

int AsebaUsbBulkRecv(unsigned char * data, unsigned char len){
	got_packet = 1;
	
restart:

	if(status == STATUS_IDLE) {
		// We are in "aseba comatible mode"
		memcpy(usb_buffer + usb_ptr, data, len);
		usb_ptr += len;
		if((usb_ptr >= *((unsigned int *) usb_buffer) + 6U) && status == STATUS_IDLE) {
			if(AsebaValidPacket()) {
				usb_ptr = 0;
			}
		}	
			
	} else {
		// We are in "Data mode" 
		int rem = len;
		if(flash_ptr + rem > 2048) 
			rem = 2048 - flash_ptr;
		memcpy(((unsigned char *)flash_buffer) + flash_ptr, data, rem);
		data += rem;
		len -= rem;
		flash_ptr += rem;
		
		if(flash_ptr == 2048) {
			int ret;
			
			if(__builtin_tbladdress(&_CODE_BASE) / 1024 <= flash_page) {
				send_ack(ASEBA_ANSWER_INVALID_VALUE);
			} else {
				if(flash_page == 0) {
					flash_in_progress = 0;
					flash_buffer[0] = 0x40000UL | (__builtin_tbloffset(&_reset) & 0xFFFE);
					flash_buffer[1] = __builtin_tblpage(&_reset);
			
					// Force the USB ALTIVT interrupt
					flash_buffer[224] =  __builtin_tbladdress(_USB1Interrupt);		
				}
				
				ret = do_flash_page(flash_page >> 6, flash_page  << 10, flash_buffer);
				
				if(ret) {
					ret = do_flash_page(flash_page >> 6, flash_page  << 10, flash_buffer);
				}
				
				if(ret) 
					send_noprog(ret - 1);
				else
					send_ack(ASEBA_ANSWER_OK);
			}
			status = STATUS_IDLE;
			usb_ptr = 0;
		}
		
		
		if(len) {
			// HACK; but works(tm);
			goto restart;
		}
	}
	return 0;
}
	
#define BOOT_VM 	0
#define BOOT_RESET	BASE_TIMEOUT

void __attribute((noreturn)) bootloader_loop(unsigned int boot_style) {
	unsigned long count = BASE_TIMEOUT + __builtin_mulus(boot_style,EXTENDED_TIMEOUT);
	flash_in_progress = 0;
	status = STATUS_IDLE;
	got_packet = 0;
	flash_ptr = 0;
	usb_ptr = 0;
	got_packet = 0;
	len = 0;
	read_status = STATUS_IDLE;
	int flags; 
	
	USBMaskInterrupts(flags);
	if(boot_style == BOOT_VM) {
		int * p = (int *) data_to_send;
		*p++ = 0;
		*p++ = 1;
		*p++ = ASEBA_MESSAGE_DISCONNECTED;
		len = 6;
		USBCDCKickTx();
	}
	USBCDCKickRx();
	USBUnmaskInterrupts(flags);
	
	do {
		
		/* Check that USB is still present */
		if(!U1OTGSTATbits.SESVD) 
			break;
		barrier();
		if(got_packet > 0) {
			got_packet = 0;
			count = BASE_TIMEOUT;
		} else if(got_packet < 0)
			break;
		
		// Hack do something better ... 	
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
		Nop();
	} while(count--);
	
	// We got a timeout ! 
	if(flash_in_progress) {
		// Arg, timeout on the USB while a flash operation is in progress ... 
		// Put ourself in "panic" mode, by writing 0x000000 at 0x200
		// This will make the bootloader reset instead of jumping to user code 
		read_page(0,0,flash_buffer);
		flash_buffer[0x100] = 0;
		flash_buffer[0] = 0x40000UL | (__builtin_tbloffset(_reset) & 0xFFFE);
		flash_buffer[1] = __builtin_tblpage(_reset);
		
		// Force the USB ALTIVT interrupt
		flash_buffer[224] = __builtin_tbladdress(_USB1Interrupt);
		
		do_flash_page(0,0,flash_buffer);
		
		USBDeviceDetach();
		goto_userspace(); // Will do a reset
	}
	
	if(got_packet == -1) {
		// Send an Ack ! 
	
	}
	goto_userspace();
}

// Huge hack ... 
static void inline init_leds(void) {
	// First, switch on Va. Wait until it's on.
#define VA _LATE2
	VA = 1;
	_TRISE2 = 0;
#define LED_CS _LATC13
	LED_CS = 0;
	_TRISC13 = 0;
	
	asm __volatile__ ("repeat  #1600 \r\n nop"); // ~100us
	
	_RP11R = 8;
	_RP12R = 7;
	SPI1STAT = 0;
	SPI1CON2 = 1;
	SPI1CON1 =  0x7B;
	
	SPI1STATbits.SPIEN = 1;
	SPI1BUF = 0x3;
	SPI1BUF = 0x80;
	SPI1BUF = 0xF;
	SPI1BUF = 0xF;
	SPI1BUF = 0xF;

	asm __volatile__ ("repeat  #80 \r\n nop"); // ~5us 
	LED_CS = 1;
	asm __volatile__ ("repeat #16 \r\n nop"); // ~1us
	LED_CS = 0;
	VA = 0; // Switch off VA ...
}

int __attribute((noreturn)) main(void) {
	unsigned long timeout = 10000000;
/* First if USB is not present, go to user code
 * Then, if _we_ initialized the usb, then wait 1 minute to get a "bootloader switch" packet
 * If not, wait a few seconds to get a bootloader valid-message
 * If nothing is RX go back to user code
 * 
 * When we write the page "0", after the the write is done, we switch to the VM ... 
 */
 	CHARGE_500MA = 1; // Temporarly enable 500mA charge ...
 						// Will be disabled right after the leds init
 						// Workaround to recover from low-level battery state.
 						// We have to enable fast charge so we have enough current
 						// To supply the leds while reseting them ...
 
 	init_leds();
 	
 	CHARGE_500MA = 0;
 
	_ALTIVT = 1;
 
 	usb_uart_init();
 	len = 0;
 
 	// Wait until we have something meaningfull on vbus sensing
 	while(!(U1OTGSTATbits.SESEND || U1OTGSTATbits.SESVD));
 
	if(U1OTGSTATbits.SESVD) {
		USBDeviceAttach();
	
		while(USBGetDeviceState() != CONFIGURED_STATE) {
			barrier();
			if(!U1OTGSTATbits.SESVD || (timeout--) == 0) {
				goto_userspace();
			}
		}
		
		bootloader_loop(BOOT_RESET);
	}
	
 	goto_userspace(); 
}

